;;;; streams-primitive.scm
;;;; Kon Lovett, Apr '09

; Copyright (C) 2007 by Philip L. Bewig of Saint Louis, Missouri, USA.  All rights
; reserved.  Permission is hereby granted, free of charge, to any person obtaining a copy of
; this software and associated documentation files (the "Software"), to deal in the Software
; without restriction, including without limitation the rights to use, copy, modify, merge,
; publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to
; whom the Software is furnished to do so, subject to the following conditions: The above
; copyright notice and this permission notice shall be included in all copies or substantial
; portions of the Software.  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
; FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
; HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
; CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
; THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(declare
  (usual-integrations)
  (fixnum)
  (inline)
  (local)
  (no-procedure-checks))

(include "chicken-primitive-object-inlines")
(include "streams-inlines")

;;;

(define-inline (%make-stream-box tag obj) (%cons tag obj))
(define-inline (%stream-box-tag box) (%car box))
(define-inline (%stream-box-value box) (%cdr box))
(define-inline (%stream-box-tag-set! box tag) (%set-car!/immediate box tag))
(define-inline (%stream-box-value-set! box val) (%set-cdr! box val))

(define-inline (%make-stream-lazy thunk) (%make-stream (%make-stream-box 'lazy thunk)))
(define-inline (%make-stream-eager obj) (%make-stream (%make-stream-box 'eager obj)))

(define-inline (%make-stream prm) (%make-structure 'stream prm))
;;(define-inline (%stream? obj) (%structure-instance? obj 'stream)) ;from "streams-inlines.scm"
(define-inline (%stream-promise strm) (%structure-ref strm 1))
(define-inline (%stream-promise-set! strm obj) (%structure-set! strm 1 obj))

(define-inline (%make-stream-pare kar kdr) (%make-structure 'stream-pare kar kdr))
(define-inline (%stream-pare? obj) (%structure-instance? obj 'stream-pare))
(define-inline (%stream-kar pare) (%structure-ref pare 1))
(define-inline (%stream-kdr pare) (%structure-ref pare 2))

(define-inline (%stream-null? strm) (%eq? (stream-force strm) (stream-force stream-null)))

(define-inline (%checked-stream-pair loc obj)
  (cond
    ((not (%stream? obj)) (error-stream loc obj 'stream) )
    ((%stream-null? obj) (error-stream-occupied loc obj 'stream) )
    (else
      (let ((val (stream-force obj)))
        (if (%stream-pare? val) val
            (error-stream-pair loc val 'stream)) ) ) ) )

(module streams-primitive (;export
  ;; SRFI 41 primitive
  stream?
  stream-null stream-null?
  (stream-cons $$stream-eager $$make-stream-pare)
  stream-pair? stream-car stream-cdr
  stream-lambda
  ;; Extras
  stream-occupied?
  ;; Common errors
  check-stream error-stream
  check-stream-occupied error-stream-occupied
  ;; WTF
  ($$stream-lazy $$make-stream-lazy)
  ($$stream-delay $$stream-eager)
  $$make-stream-lazy
  $$stream-eager
  $$make-stream-pare)

(import scheme chicken
  (only type-checks define-check+error-type)
  (only type-errors define-error-type))

(require-library type-checks type-errors)

;;;

(define-check+error-type stream)
(define-check+error-type stream-occupied)
(define-error-type stream-pair)

;;;

(define (stream? obj) (%stream? obj))

(define ($$make-stream-lazy thunk) (%make-stream-lazy thunk))

(define-syntax $$stream-lazy
  (syntax-rules ()
    ((_ EXPR) ($$make-stream-lazy (lambda () EXPR)) ) ) )

(define ($$stream-eager obj) (%make-stream-eager obj))

(define-syntax $$stream-delay
  (syntax-rules ()
    ((_ EXPR) ($$stream-lazy ($$stream-eager EXPR)) ) ) )

(define ($$make-stream-pare kar kdr) (%make-stream-pare kar kdr))

;;;

(define (stream-force promise)
  (let ((content (%stream-promise promise)))
    (case (%stream-box-tag content)
      ((eager)
        (%stream-box-value content) )
      ((lazy)
        (let* ((promise* ((%stream-box-value content)))
               (content  (%stream-promise promise)))
          (unless (%eq? 'eager (%stream-box-tag content))
            (let ((prm (%stream-promise promise*)))
              (%stream-box-tag-set! content (%stream-box-tag prm))
              (%stream-box-value-set! content (%stream-box-value prm)) )
            (%stream-promise-set! promise* content) )
         (stream-force promise) ) ) ) ) )

(define stream-null ($$stream-delay (%cons 'stream 'null)))

(define (stream-null? obj) (and (%stream? obj) (%stream-null? obj)))
(define (stream-occupied? obj) (and (%stream? obj) (not (%stream-null? obj))))

(define-syntax stream-cons
  (syntax-rules ()
    ((_ EXPR STRM)
     ($$stream-eager ($$make-stream-pare ($$stream-delay EXPR) ($$stream-lazy STRM))) ) ) )

(define (stream-pair? obj) (and (%stream? obj) (%stream-pare? (stream-force obj))))

(define (stream-car streem)
  (stream-force (%stream-kar (%checked-stream-pair 'stream-car streem))) )

(define (stream-cdr streem)
  (%stream-kdr (%checked-stream-pair 'stream-cdr streem)) )

(define-syntax stream-lambda
  (syntax-rules ()
    ((_ FORMALS BODY0 BODY1 ...)
     (lambda FORMALS ($$stream-lazy (let () BODY0 BODY1 ...))) ) ) )

) ;module streams-primitive
